# -*- coding: utf-8 -*-
"""
    address.py

    :copyright: (c) 2014 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
from trytond.pool import PoolMeta

__metaclass__ = PoolMeta
__all__ = ["Address"]


class Address:
    __name__ = "party.address"

    def serialize(self, purpose=None):
        """
        Address serialization for the purpose of POS
        """
        if purpose == 'pos':
            return {
                'id': self.id,
                'name': self.name,
                'full_address': self.full_address,
            }
        elif hasattr(super(Address, self), 'serialize'):
            return super(Address, self).serialize(purpose)  # pragma: no cover



# -*- coding: utf-8 -*-
"""
    __init__.py

    :copyright: (c) 2014 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
from trytond.pool import Pool

from sale import Sale, SaleChannel, SaleLine, SaleConfiguration
from address import Address
from shipment import ShipmentOut, ShipmentOutReturn


def register():
    Pool.register(
        SaleChannel,
        Sale,
        SaleLine,
        ShipmentOut,
        ShipmentOutReturn,
        Address,
        SaleConfiguration,
        module='pos', type_='model'
    )



# -*- coding: utf-8 -*-
"""
    tests/test_sale.py

    :copyright: (C) 2014 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
import sys
import os
import unittest
import datetime
from decimal import Decimal
from dateutil.relativedelta import relativedelta

import trytond.tests.test_tryton
from trytond.tests.test_tryton import POOL, USER, DB_NAME, CONTEXT
from trytond.transaction import Transaction
from trytond.exceptions import UserError

DIR = os.path.abspath(os.path.normpath(os.path.join(
    __file__, '..', '..', '..', '..', '..', 'trytond'
)))
if os.path.isdir(DIR):
    sys.path.insert(0, os.path.dirname(DIR))


class TestSale(unittest.TestCase):
    '''
    Sale Test Case for lie-nielsen module.
    '''

    def setUp(self):
        """
        Set up data used in the tests.
        this method is called before each test function execution.
        """
        trytond.tests.test_tryton.install_module('pos')
        self.Company = POOL.get('company.company')
        self.Party = POOL.get('party.party')
        self.Address = POOL.get('party.address')
        self.Currency = POOL.get('currency.currency')
        self.User = POOL.get('res.user')
        self.Location = POOL.get('stock.location')
        self.PriceList = POOL.get('product.price_list')
        self.PaymentTerm = POOL.get('account.invoice.payment_term')
        self.Sequence = POOL.get('ir.sequence')
        self.Sale = POOL.get('sale.sale')
        self.SaleLine = POOL.get('sale.line')
        self.Channel = POOL.get('sale.channel')
        self.Product = POOL.get('product.template')
        self.SaleConfiguration = POOL.get('sale.configuration')
        self.Invoice = POOL.get('account.invoice')
        self.InvoiceLine = POOL.get('account.invoice.line')

    def _create_product_category(self, name):
        """
        Creates a product category

        Name is mandatory while other value may be provided as keyword
        arguments

        :param name: Name of the product category
        """
        Category = POOL.get('product.category')

        return Category.create([{
            'name': name,
        }])

    def _create_product_template(self, name, vlist, uom=u'Unit'):
        """
        Create a product template with products and return its ID

        :param name: Name of the product
        :param vlist: List of dictionaries of values to create
        :param uom: Note it is the name of UOM (not symbol or code)
        """
        ProductTemplate = POOL.get('product.template')
        Uom = POOL.get('product.uom')

        for values in vlist:
            values['name'] = name
            values['default_uom'], = Uom.search([('name', '=', uom)], limit=1)
            values['sale_uom'], = Uom.search([('name', '=', uom)], limit=1)
            values['products'] = [
                ('create', [{}])
            ]
        return ProductTemplate.create(vlist)

    def _create_fiscal_year(self, date=None, company=None):
        """
        Creates a fiscal year and requried sequences
        """
        FiscalYear = POOL.get('account.fiscalyear')
        Sequence = POOL.get('ir.sequence')
        SequenceStrict = POOL.get('ir.sequence.strict')
        Company = POOL.get('company.company')

        if date is None:
            date = datetime.date.today()

        if company is None:
            company, = Company.search([], limit=1)

        invoice_sequence, = SequenceStrict.create([{
            'name': '%s' % date.year,
            'code': 'account.invoice',
            'company': company,
        }])
        fiscal_year, = FiscalYear.create([{
            'name': '%s' % date.year,
            'start_date': date + relativedelta(month=1, day=1),
            'end_date': date + relativedelta(month=12, day=31),
            'company': company,
            'post_move_sequence': Sequence.create([{
                'name': '%s' % date.year,
                'code': 'account.move',
                'company': company,
            }])[0],
            'out_invoice_sequence': invoice_sequence,
            'in_invoice_sequence': invoice_sequence,
            'out_credit_note_sequence': invoice_sequence,
            'in_credit_note_sequence': invoice_sequence,
        }])
        FiscalYear.create_period([fiscal_year])
        return fiscal_year

    def _create_coa_minimal(self, company):
        """Create a minimal chart of accounts
        """
        AccountTemplate = POOL.get('account.account.template')
        Account = POOL.get('account.account')

        account_create_chart = POOL.get(
            'account.create_chart', type="wizard")

        account_template, = AccountTemplate.search(
            [('parent', '=', None)]
        )

        session_id, _, _ = account_create_chart.create()
        create_chart = account_create_chart(session_id)
        create_chart.account.account_template = account_template
        create_chart.account.company = company
        create_chart.transition_create_account()

        receivable, = Account.search([
            ('kind', '=', 'receivable'),
            ('company', '=', company),
        ])
        payable, = Account.search([
            ('kind', '=', 'payable'),
            ('company', '=', company),
        ])
        create_chart.properties.company = company
        create_chart.properties.account_receivable = receivable
        create_chart.properties.account_payable = payable
        create_chart.transition_create_properties()

    def _get_account_by_kind(self, kind, company=None, silent=True):
        """Returns an account with given spec

        :param kind: receivable/payable/expense/revenue
        :param silent: dont raise error if account is not found
        """
        Account = POOL.get('account.account')
        Company = POOL.get('company.company')

        if company is None:
            company, = Company.search([], limit=1)

        accounts = Account.search([
            ('kind', '=', kind),
            ('company', '=', company)
        ], limit=1)
        if not accounts and not silent:
            raise Exception("Account not found")
        return accounts[0] if accounts else None

    def _create_payment_term(self):
        """Create a simple payment term with all advance
        """
        PaymentTerm = POOL.get('account.invoice.payment_term')

        return PaymentTerm.create([{
            'name': 'Direct',
            'lines': [('create', [{'type': 'remainder'}])]
        }])

    def _create_pricelists(self):
        """
        Create the pricelists
        """
        # Setup the pricelists
        self.party_pl_margin = Decimal('1.10')
        self.guest_pl_margin = Decimal('1.20')
        user_price_list, = self.PriceList.create([{
            'name': 'PL 1',
            'company': self.company.id,
            'lines': [
                ('create', [{
                    'formula': 'unit_price * %s' % self.party_pl_margin
                }])
            ],
        }])
        guest_price_list, = self.PriceList.create([{
            'name': 'PL 2',
            'company': self.company.id,
            'lines': [
                ('create', [{
                    'formula': 'unit_price * %s' % self.guest_pl_margin
                }])
            ],
        }])
        return guest_price_list.id, user_price_list.id

    def setup_defaults(self):
        """
        Setup Defaults
        """
        Uom = POOL.get('product.uom')
        AccountTax = POOL.get('account.tax')
        Account = POOL.get('account.account')
        Inventory = POOL.get('stock.inventory')

        self.usd, = self.Currency.create([{
            'name': 'US Dollar',
            'code': 'USD',
            'symbol': '$',
        }])

        Country = POOL.get('country.country')
        self.country, = Country.create([{
            'name': 'United States of America',
            'code': 'US',
        }])

        Subdivision = POOL.get('country.subdivision')
        self.subdivision, = Subdivision.create([{
            'country': self.country.id,
            'name': 'California',
            'code': 'CA',
            'type': 'state',
        }])

        self.uom, = Uom.search([('symbol', '=', 'u')], limit=1)

        with Transaction().set_user(0):
            self.party, = self.Party.create([{
                'name': 'Openlabs',
                'addresses': [('create', [{
                    'name': 'Lie Nielsen',
                    'city': 'Los Angeles',
                    'country': self.country.id,
                    'subdivision': self.subdivision.id,
                }])],
            }])
            self.anonymous_customer, = self.Party.create([{
                'name': 'Anonymous Customer Party'
            }])
            self.address, = self.Address.create([{
                'party': self.anonymous_customer,
                'name': 'Jon Doe\'s Address'
            }])
            self.company, = self.Company.create([{
                'party': self.party.id,
                'currency': self.usd
            }])
            user = self.User(USER)
            self.User.write([user], {
                'main_company': self.company.id,
                'company': self.company.id,
            })

        with Transaction().set_context(company=self.company.id):
            # Create Fiscal Year
            self._create_fiscal_year(company=self.company.id)
            # Create Chart of Accounts
            self._create_coa_minimal(company=self.company.id)
            # Create a payment term
            self._create_payment_term()

            sequence, = self.Sequence.search([
                ('code', '=', 'sale.sale'),
            ], limit=1)
            warehouse, = self.Location.search([
                ('code', '=', 'WH'),
            ])
            self.payment_term, = self.PaymentTerm.create([{
                'name': 'Payment term',
                'lines': [
                    ('create', [{
                        'sequence': 0,
                        'type': 'remainder',
                        'days': 0,
                        'months': 0,
                        'weeks': 0,
                    }])]
            }])
            price_list, = self.PriceList.create([{
                'name': 'PL 1',
                'company': self.company.id,
                'lines': [
                    ('create', [{
                        'formula': 'unit_price'
                    }])
                ],
            }])

            self.channel, = self.Channel.create([{
                'name': 'Channel',
                'company': self.company.id,
                'source': 'manual',
                'currency': self.usd.id,
                'anonymous_customer': self.anonymous_customer.id,
                'warehouse': warehouse.id,
                'ship_from_warehouse': warehouse.id,
                'price_list': price_list.id,
                'payment_term': self.payment_term.id,
                'invoice_method': 'order',
                'shipment_method': 'manual',
            }])
            self.channel1, = self.Channel.create([{
                'name': 'Channel 1',
                'company': self.company.id,
                'source': 'manual',
                'currency': self.usd.id,
                'anonymous_customer': self.anonymous_customer.id,
                'warehouse': warehouse.id,
                'ship_from_warehouse': warehouse.id,
                'price_list': price_list.id,
                'payment_term': self.payment_term.id,
                'invoice_method': 'order',
                'shipment_method': 'manual',
            }])
            user = self.User(USER)
            self.User.write([user], {
                'create_channels': [('add', [self.channel, self.channel1])],
                'current_channel': self.channel.id,
            })
            account = Account.search([('name', '=', 'Main Tax')])[0]
            tax, = AccountTax.create([{
                'name': 'Test Tax',
                'description': 'Test Tax',
                'rate': Decimal('0.10'),
                'invoice_account': account,
                'credit_note_account': account,
            }])

        self.category, = self._create_product_category(
            'Category'
        )

        # Create product templates with products
        self.template1, = self._create_product_template(
            'product-1',
            [{
                'category': self.category.id,
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('10'),
                'cost_price': Decimal('5'),
                'account_expense': self._get_account_by_kind('expense').id,
                'account_revenue': self._get_account_by_kind('revenue').id,
            }]
        )
        self.template2, = self._create_product_template(
            'product-2',
            [{
                'category': self.category.id,
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('15'),
                'cost_price': Decimal('5'),
                'account_expense': self._get_account_by_kind('expense').id,
                'account_revenue': self._get_account_by_kind('revenue').id,
            }]
        )
        self.template3, = self._create_product_template(
            'product-3',
            [{
                'category': self.category.id,
                'type': 'goods',
                'salable': True,
                'list_price': Decimal('15'),
                'cost_price': Decimal('5'),
                'account_expense': self._get_account_by_kind('expense').id,
                'account_revenue': self._get_account_by_kind('revenue').id,
                'customer_taxes': [('add', [tax])]
            }]
        )
        self.template4, = self._create_product_template(
            'product-4',
            [{
                'category': self.category.id,
                'type': 'service',
                'salable': True,
                'list_price': Decimal('10'),
                'cost_price': Decimal('5'),
                'account_expense': self._get_account_by_kind('expense').id,
                'account_revenue': self._get_account_by_kind('revenue').id,
            }]
        )
        self.product1 = self.template1.products[0]
        self.product2 = self.template2.products[0]
        self.product3 = self.template3.products[0]
        self.product4 = self.template4.products[0]

        inventory, = Inventory.create([{
            'location': warehouse.storage_location,
            'company': self.company.id,
            'lines': [('create', [{
                'product': self.product1,
                'quantity': 20,
            }])]
        }])
        Inventory.confirm([inventory])

    def test_0010_test_sale(self):
        """
        Sale model is not broken
        """
        with Transaction().start(DB_NAME, USER, CONTEXT):
            self.setup_defaults()

            with Transaction().set_context(use_anonymous_customer=True):
                sale, = self.Sale.create([{
                    'currency': self.usd.id,
                }])

            with Transaction().set_context(
                    company=self.company.id, channel=self.channel.id
            ):
                sale.pos_add_product([self.product1.id], 1)
                self.assertEqual(len(sale.lines), 1)
                sale.pos_add_product([self.product1.id], 2)
                self.assertEqual(len(sale.lines), 1)
                self.assertEqual(sale.lines[0].quantity, 2)

                rv = sale.pos_add_product([self.product2.id], 2)
                self.assertEqual(len(sale.lines), 2)

            self.assertEqual(len(rv['sale']['lines']), 2)

    def test_0020_test_delivery_mode_on_adding(self):
        """
        Ensure that delivery mode is respected when added to cart
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            with Transaction().set_context(use_anonymous_customer=True):
                sale, = self.Sale.create([{
                    'currency': self.usd.id,
                }])

            with Transaction().set_context(
                    company=self.company.id, channel=self.channel.id
            ):
                rv = sale.pos_add_product([self.product1.id], 1)

                # By default the lines are picked
                self.assertEqual(len(rv['sale']['lines']), 1)
                self.assertEqual(
                    rv['sale']['lines'][0]['delivery_mode'], 'pick_up'
                )
                self.assertEqual(rv['sale']['lines'][0]['quantity'], 1)

                # Add another line, but with explicit delivery_mode
                with Transaction().set_context(delivery_mode='pick_up'):
                    rv = sale.pos_add_product([self.product1.id], 2)
                self.assertEqual(len(rv['sale']['lines']), 1)
                self.assertEqual(
                    rv['sale']['lines'][0]['delivery_mode'], 'pick_up'
                )
                self.assertEqual(rv['sale']['lines'][0]['quantity'], 2)

                # Add a ship line of same product
                with Transaction().set_context(delivery_mode='ship'):
                    rv = sale.pos_add_product([self.product1.id], 1)
                    self.assertEqual(len(rv['sale']['lines']), 2)

                    for pick_line in filter(
                            lambda l: l['delivery_mode'] == 'pick_up',
                            rv['sale']['lines']):
                        # From the previous addition
                        self.assertEqual(pick_line['delivery_mode'], 'pick_up')
                        self.assertEqual(pick_line['quantity'], 2)
                        break
                    else:
                        self.fail('Expected to find pick up line, but did not')

                    for ship_line in filter(
                            lambda l: l['delivery_mode'] == 'ship',
                            rv['sale']['lines']):
                        self.assertEqual(ship_line['delivery_mode'], 'ship')
                        self.assertEqual(ship_line['quantity'], 1)
                        break
                    else:
                        self.fail('Expected to find line, but did not')

    def test_0022_test_update_delivery_mode(self):
        """
        Update delivery mode of saleLine
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            with Transaction().set_context(use_anonymous_customer=True):
                sale, = self.Sale.create([{
                    'currency': self.usd.id,
                }])

            with Transaction().set_context(
                    company=self.company.id, channel=self.channel.id
            ):
                rv = sale.pos_add_product([self.product1.id], 1)

                # By default the lines are picked
                self.assertEqual(len(rv['sale']['lines']), 1)
                self.assertEqual(
                    rv['sale']['lines'][0]['delivery_mode'], 'pick_up'
                )
                self.assertEqual(rv['sale']['lines'][0]['quantity'], 1)

            # Update delivery_mode in sale line
            with Transaction().set_context(
                    delivery_mode='ship',
                    sale_line=rv['updated_lines'][0]
            ):
                rv = sale.pos_add_product([self.product1.id], 2)

            self.assertEqual(len(rv['sale']['lines']), 1)
            self.assertEqual(rv['sale']['lines'][0]['delivery_mode'], 'ship')
            self.assertEqual(rv['sale']['lines'][0]['quantity'], 2)

            # Change product and provide saleLine
            with Transaction().set_context(
                    delivery_mode='ship',
                    sale_line=rv['updated_lines'][0]
            ):
                rv = sale.pos_add_product([self.product2.id], 2)

            self.assertEqual(len(rv['sale']['lines']), 1)
            # Product should not change
            self.assertEqual(
                rv['sale']['lines'][0]['product']['id'], self.product1.id
            )
            self.assertEqual(rv['sale']['lines'][0]['delivery_mode'], 'ship')
            self.assertEqual(rv['sale']['lines'][0]['quantity'], 2)

    def test_0025_add_taxes_on_line(self):
        """
        Add a line that woudl add taxes and check that it works
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            with Transaction().set_context(use_anonymous_customer=True):
                sale, = self.Sale.create([{
                    'currency': self.usd.id,
                }])

            with Transaction().set_context(
                    company=self.company.id, channel=self.channel.id
            ):
                rv = sale.pos_add_product([self.product1.id], 1)

                # add a product which does not have taxes
                self.assertEqual(len(rv['sale']['lines']), 1)
                sale_line = self.SaleLine(rv['updated_lines'][0])
                self.assertFalse(sale_line.taxes)
                self.assertEqual(rv['sale']['tax_amount'], 0)

                rv = sale.pos_add_product([self.product3.id], 1)

                # add a product which does not have taxes
                self.assertEqual(len(rv['sale']['lines']), 2)
                sale_line = self.SaleLine(rv['updated_lines'][0])
                self.assertEqual(rv['sale']['tax_amount'], Decimal('1.5'))

                # Please make that two ;)
                rv = sale.pos_add_product([self.product3.id], 2)

                # add a product which does not have taxes
                self.assertEqual(len(rv['sale']['lines']), 2)
                sale_line = self.SaleLine(rv['updated_lines'][0])
                self.assertEqual(rv['sale']['tax_amount'], Decimal('3'))

    def test_0030_serialization_fallback(self):
        """
        Ensure that serialization for other purposes still work
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            with Transaction().set_context(use_anonymous_customer=True):
                sale, = self.Sale.create([{
                    'currency': self.usd.id,
                }])

            with Transaction().set_context(
                    company=self.company.id, channel=self.channel.id):
                rv = sale.pos_add_product([self.product1.id], 1)
                sale_line = self.SaleLine(rv['updated_lines'][0])

            # Serialize sale
            sale.serialize()
            sale_line.serialize()

    def test_0035_sale_pos_serialization(self):
        """
        Serialize sale for pos
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            with Transaction().set_context(use_anonymous_customer=True):
                sale, = self.Sale.create([{
                    'currency': self.usd.id,
                    'invoice_address': self.address,
                    'shipment_address': self.address,
                }])

            with Transaction().set_context(
                    company=self.company.id, channel=self.channel.id):
                sale.pos_add_product([self.product1.id], 1)

            # Serialize sale for pos
            rv = sale.pos_serialize()
            self.assertEqual(rv['total_amount'], sale.total_amount)
            self.assertEqual(rv['tax_amount'], sale.tax_amount)
            self.assertEqual(len(rv['lines']), 1)

    def test_0040_default_delivery_mode(self):
        """
        Test default delivery_mode for saleLine
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            with Transaction().set_context(
                use_anonymous_customer=True, channel=self.channel.id
            ):
                sale, = self.Sale.create([{
                    'currency': self.usd.id,
                }])

            self.assertEqual(sale.channel.delivery_mode, 'ship')
            with Transaction().set_context(
                company=self.company.id, channel=self.channel.id,
                current_sale_channel=self.channel.id
            ):
                sale_line, = self.SaleLine.create([{
                    'sale': sale.id,
                    'product': self.product1.id,
                    'description': 'test product',
                    'quantity': 1,
                    'unit': self.product1.default_uom.id,
                    'unit_price': Decimal('10'),
                }])
                self.assertEqual(
                    sale_line.delivery_mode, self.channel.delivery_mode
                )

            with Transaction().set_user(0):
                with Transaction().set_context(
                    company=self.company.id, channel=None
                ):
                    new_sale_line, = self.SaleLine.create([{
                        'sale': sale.id,
                        'product': self.product4.id,
                        'description': 'test service product',
                        'quantity': 1,
                        'unit': self.product4.default_uom.id,
                        'unit_price': Decimal('10'),
                    }])
                    self.assertIsNone(new_sale_line.delivery_mode)

            # Test if sale line's product type is goods
            self.assertTrue(sale_line.product_type_is_goods)
            self.assertFalse(new_sale_line.product_type_is_goods)

    def test_0120_ship_pick_diff_warehouse(self):
        """
        Ensure that ship_from_warehouse is used for back orders while orders
        are picked from the channel's warehouse
        """
        Location = POOL.get('stock.location')
        Channel = POOL.get('sale.channel')
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            ship_from_wh, = Location.copy([self.channel.warehouse])

            # Set that as the new ship_from warehouse
            Channel.write([self.channel], {
                'ship_from_warehouse': ship_from_wh.id}
            )

            with Transaction().set_context({
                    'company': self.company.id,
                    'channel': self.channel.id,
                    'channels': [self.channel.id], }):
                # Now create an order
                sale, = self.Sale.create([{
                    'reference': 'Test Sale',
                    'payment_term': self.payment_term,
                    'currency': self.company.currency.id,
                    'party': self.party.id,
                    'invoice_address': self.party.addresses[0].id,
                    'shipment_address': self.party.addresses[0].id,
                    'sale_date': Date.today(),
                    'company': self.company.id,

                    # keep invoicing out of the way for this test's sake
                    'invoice_method': 'manual',
                    'shipment_method': 'order',

                    # Explicitly specify the channel
                    'channel': Channel(self.channel).id,
                }])
                self.SaleLine.create([{
                    'sale': sale,
                    'type': 'line',
                    'quantity': 2,
                    'delivery_mode': 'pick_up',
                    'unit': self.uom,
                    'unit_price': 20000,
                    'description': 'Picked Item',
                    'product': self.product1.id
                }, {
                    'sale': sale,
                    'type': 'line',
                    'quantity': 2,
                    'delivery_mode': 'ship',
                    'unit': self.uom,
                    'unit_price': 20000,
                    'description': 'Shipped Item',
                    'product': self.product1.id
                }])

                # Quote, Confirm and Process the order
                self.Sale.quote([sale])
                self.Sale.confirm([sale])
                self.Sale.process([sale])

                self.assertEqual(len(sale.shipments), 2)
                for shipment in sale.shipments:
                    if shipment.delivery_mode == 'pick_up':
                        self.assertEqual(shipment.state, 'done')
                        self.assertEqual(
                            shipment.warehouse, self.channel.warehouse
                        )
                    elif shipment.delivery_mode == 'ship':
                        self.assertEqual(shipment.state, 'waiting')
                        self.assertEqual(
                            shipment.warehouse, self.channel.ship_from_warehouse
                        )
                    else:
                        self.fail('Invalid delivery mode')

    def test_1010_delivery_method_2shipping_case_1(self):
        """
        Ensure shipment method is respected by sale order processing

        Case 1: Ship only order
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'reference': 'Test Sale',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,

                # keep invoicing out of the way for this test's sake
                'invoice_method': 'manual',
                'shipment_method': 'order',
            }])
            sale_line, = self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'ship',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Test description',
                'product': self.product1.id
            }])

            # Quote, Confirm and Process the order
            self.Sale.quote([sale])
            self.Sale.confirm([sale])
            self.Sale.process([sale])

            self.assertEqual(len(sale.shipments), 1)
            self.assertEqual(sale.shipments[0].delivery_mode, 'ship')
            self.assertEqual(sale.shipments[0].state, 'waiting')

    def test_1020_delivery_method_2shipping_case_2(self):
        """
        Ensure shipment method is respected by sale order processing

        Case 2: Pick up only order
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'reference': 'Test Sale',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,

                # keep invoicing out of the way for this test's sake
                'invoice_method': 'manual',
                'shipment_method': 'order',
            }])
            sale_line, = self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'pick_up',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Test description',
                'product': self.product1.id
            }])

            # Quote, Confirm and Process the order
            self.Sale.quote([sale])
            self.Sale.confirm([sale])
            self.Sale.process([sale])

            self.assertEqual(len(sale.shipments), 1)
            self.assertEqual(sale.shipments[0].delivery_mode, 'pick_up')
            self.assertEqual(sale.shipments[0].state, 'done')

    def test_1030_delivery_method_2shipping_case_3(self):
        """
        Ensure shipment method is respected by sale order processing

        Case 2: Pick up + Ship order for same item
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'reference': 'Test Sale',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,

                # keep invoicing out of the way for this test's sake
                'invoice_method': 'manual',
                'shipment_method': 'order',
            }])
            self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'pick_up',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Picked Item',
                'product': self.product1.id
            }, {
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'ship',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Shipped Item',
                'product': self.product1.id
            }])

            # Quote, Confirm and Process the order
            self.Sale.quote([sale])
            self.Sale.confirm([sale])
            self.Sale.process([sale])

            self.assertEqual(len(sale.shipments), 2)
            for shipment in sale.shipments:
                if shipment.delivery_mode == 'pick_up':
                    self.assertEqual(shipment.state, 'done')
                elif shipment.delivery_mode == 'ship':
                    self.assertEqual(shipment.state, 'waiting')
                else:
                    self.fail('Invalid delivery mode')

    def test_1040_delivery_method_2shipping_case_4(self):
        """
        Manual shipping should just go ahead without messing with new
        workflow
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'reference': 'Test Sale',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,

                # keep invoicing out of the way for this test's sake
                'invoice_method': 'manual',
                'shipment_method': 'manual',
            }])
            self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'pick_up',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Picked Item',
                'product': self.product1.id
            }, {
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'ship',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Shipped Item',
                'product': self.product1.id
            }])

            # Quote, Confirm and Process the order
            self.Sale.quote([sale])
            self.Sale.confirm([sale])
            self.Sale.process([sale])
            self.assertEqual(len(sale.shipments), 0)

    def test_1050_delivery_method_2shipping_case_5(self):
        """
        Return Shipment
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'reference': 'Test Sale',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,

                # keep invoicing out of the way for this test's sake
                'invoice_method': 'manual',
                'shipment_method': 'order',
            }])
            self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': -2,
                'delivery_mode': 'pick_up',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Picked Item',
                'product': self.product1.id
            }])

            # Quote, Confirm and Process the order
            self.Sale.quote([sale])
            self.Sale.confirm([sale])
            self.Sale.process([sale])
            self.assertEqual(len(sale.shipments), 0)
            self.assertEqual(len(sale.shipment_returns), 1)

    def test_1090_default_delivery_methods(self):
        """
        Defaults should be to ship products
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            ShipmentOut = POOL.get('stock.shipment.out')
            ShipmentOutReturn = POOL.get('stock.shipment.out.return')

            with Transaction().set_context({'company': self.company.id}):
                shipment, = ShipmentOut.create([{
                    'customer': self.party.id,
                    'delivery_address': self.party.addresses[0].id,
                }])
                self.assertEqual(shipment.delivery_mode, 'ship')

                shipment_return, = ShipmentOutReturn.create([{
                    'customer': self.party.id,
                    'delivery_address': self.party.addresses[0].id,
                }])
                self.assertEqual(shipment_return.delivery_mode, 'ship')

    def test_1110_shipment_invoice_case_1(self):
        """
        Ensure that a posted invoice is created when a picked up order is
        processed.

        For the ship order, since nothing has been shipped, there should be
        no invoices
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'reference': 'Test Sale',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,

                # keep invoicing out of the way for this test's sake
                'invoice_method': 'shipment',
                'shipment_method': 'order',
            }])
            sale_line, = self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'ship',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Test description',
                'product': self.product1.id
            }])

            # Quote, Confirm and Process the order
            self.Sale.quote([sale])
            self.Sale.confirm([sale])
            self.Sale.process([sale])

            self.assertEqual(len(sale.shipments), 1)
            self.assertEqual(sale.shipments[0].delivery_mode, 'ship')
            self.assertEqual(sale.shipments[0].state, 'waiting')

            self.assertEqual(len(sale.invoices), 0)

    def test_1120_shipment_invoice_case_2(self):
        """
        Ensure that a posted invoice is created when a picked up order is
        processed
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'reference': 'Test Sale',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,

                # keep invoicing out of the way for this test's sake
                'invoice_method': 'shipment',
                'shipment_method': 'order',
            }])
            sale_line, = self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'pick_up',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Test description',
                'product': self.product1.id
            }])

            with Transaction().set_context({'company': self.company.id}):
                # Quote, Confirm and Process the order
                self.Sale.quote([sale])
                self.Sale.confirm([sale])
                self.Sale.process([sale])

                self.assertEqual(len(sale.shipments), 1)
                self.assertEqual(sale.shipments[0].delivery_mode, 'pick_up')
                self.assertEqual(sale.shipments[0].state, 'done')

                self.assertEqual(len(sale.invoices), 1)
                self.assertEqual(sale.invoices[0].state, 'posted')

    def test_1130_delivery_method_2shipping_case_3(self):
        """
        Ensure shipment method is respected by sale order processing.
        Ensures that there is only one invoice created and that it is posted.

        Case 3: Pick up + Ship order for same item
        """
        Date = POOL.get('ir.date')
        Shipment = POOL.get('stock.shipment.out')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'reference': 'Test Sale',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,

                # keep invoicing out of the way for this test's sake
                'invoice_method': 'shipment',
                'shipment_method': 'order',
            }])
            self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'pick_up',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Picked Item',
                'product': self.product1.id
            }, {
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'ship',
                'unit': self.uom,
                'unit_price': 20000,
                'description': 'Shipped Item',
                'product': self.product1.id
            }])

            with Transaction().set_context({'company': self.company.id}):
                # Quote, Confirm and Process the order
                self.Sale.quote([sale])
                self.Sale.confirm([sale])
                self.Sale.process([sale])

            self.assertEqual(len(sale.shipments), 2)
            for shipment in sale.shipments:
                if shipment.delivery_mode == 'pick_up':
                    self.assertEqual(shipment.state, 'done')
                elif shipment.delivery_mode == 'ship':
                    self.assertEqual(shipment.state, 'waiting')
                    delivery_shipment = shipment        # used later in test
                else:
                    self.fail('Invalid delivery mode')

            self.assertEqual(len(sale.invoices), 1)
            self.assertEqual(sale.invoices[0].state, 'posted')

            with Transaction().set_context({'company': self.company.id}):
                # Now process the delivered shipment as if its been shipped
                Shipment.assign_force([delivery_shipment])
                Shipment.pack([delivery_shipment])
                Shipment.done([delivery_shipment])

                self.assertEqual(len(sale.invoices), 2)
                self.assertEqual(sale.invoices[0].state, 'posted')
                self.assertEqual(sale.invoices[1].state, 'posted')

    def test_1140_serialize_recent_sales(self):
        """
        Test that sale order which are recently updated or create are on top.
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            with Transaction().set_context(current_channel=self.channel.id):
                sale1, = self.Sale.create([{
                    'reference': 'Test Sale 1',
                    'payment_term': self.payment_term,
                    'currency': self.company.currency.id,
                    'party': self.party.id,
                    'invoice_address': self.party.addresses[0].id,
                    'shipment_address': self.party.addresses[0].id,
                    'sale_date': Date.today(),
                    'company': self.company.id,
                }])
                sale2, = self.Sale.create([{
                    'reference': 'Test Sale 2',
                    'payment_term': self.payment_term,
                    'currency': self.company.currency.id,
                    'party': self.party.id,
                    'invoice_address': self.party.addresses[0].id,
                    'shipment_address': self.party.addresses[0].id,
                    'sale_date': Date.today(),
                    'company': self.company.id,
                }])

                sale3, = self.Sale.create([{
                    'reference': 'Test Sale 3',
                    'payment_term': self.payment_term,
                    'currency': self.company.currency.id,
                    'party': self.party.id,
                    'invoice_address': self.party.addresses[0].id,
                    'shipment_address': self.party.addresses[0].id,
                    'sale_date': Date.today(),
                    'company': self.company.id,
                }])
                sale4, = self.Sale.create([{
                    'reference': 'Test Sale 4',
                    'payment_term': self.payment_term,
                    'currency': self.company.currency.id,
                    'party': self.party.id,
                    'invoice_address': self.party.addresses[0].id,
                    'shipment_address': self.party.addresses[0].id,
                    'sale_date': Date.today(),
                    'company': self.company.id,
                }])

                saleLine1, = self.SaleLine.create([{
                    'sale': sale1,
                    'type': 'line',
                    'quantity': 2,
                    'delivery_mode': 'pick_up',
                    'unit': self.uom,
                    'unit_price': 20000,
                    'description': 'Picked Item',
                    'product': self.product1.id
                }])
                saleLine2, = self.SaleLine.create([{
                    'sale': sale2,
                    'type': 'line',
                    'quantity': 2,
                    'delivery_mode': 'pick_up',
                    'unit': self.uom,
                    'unit_price': 20000,
                    'description': 'Picked Item',
                    'product': self.product1.id
                }])
                saleLine3, = self.SaleLine.create([{
                    'sale': sale3,
                    'type': 'line',
                    'quantity': 2,
                    'delivery_mode': 'pick_up',
                    'unit': self.uom,
                    'unit_price': 20000,
                    'description': 'Picked Item',
                    'product': self.product1.id
                }])
                saleLine4, = self.SaleLine.create([{
                    'sale': sale4,
                    'type': 'line',
                    'quantity': 2,
                    'delivery_mode': 'pick_up',
                    'unit': self.uom,
                    'unit_price': 20000,
                    'description': 'Picked Item',
                    'product': self.product1.id
                }])

                rv = self.Sale.get_recent_sales()
                self.assertEqual(len(rv), 4)

                # Test serialized data
                self.assertIn('id', rv[0])
                self.assertIn('party', rv[0])
                self.assertIn('total_amount', rv[0])
                self.assertIn('create_date', rv[0])

    def test_1150_round_off_case_1(self):
        """
        Test round off in sale and invoice.
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            product, = self.Product.create([{
                'name': 'Test product',
                'list_price': 200,
                'cost_price': 200,
                'default_uom': self.uom,
                'salable': True,
                'sale_uom': self.uom,
                'account_expense': self._get_account_by_kind('expense').id,
                'account_revenue': self._get_account_by_kind('revenue').id,
                'products': [('create', [
                    {}
                ])]
            }])
            sale, = self.Sale.create([{
                'reference': 'Test Sale 1',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'company': self.company.id,
                'lines': [('create', [
                    {
                        'type': 'line',
                        'quantity': 1,
                        'product': product.products[0].id,
                        'unit': self.uom,
                        'unit_price': Decimal(200.25),
                        'description': 'sale line',
                    }
                ])]
            }])

            with Transaction().set_context(company=self.company.id):
                self.Sale.round_down_total([sale])
                self.assertEqual(len(sale.lines), 2)

                round_off_line, = self.SaleLine.search([
                    ('is_round_off', '=', True)
                ])

                # There should be a new line of type 'roundoff'
                self.assertIsNotNone(round_off_line)

                # Total order price 200.25 should have been rounded down to 200
                self.assertEqual(sale.total_amount, 200)
                # Difference after rounding down should be created as
                # roundoff line.
                self.assertEqual(round_off_line.unit_price, 0.25)
                self.assertEqual(round_off_line.quantity, -1)
                self.assertEqual(round_off_line.amount, -0.25)

                self.SaleLine.create([
                    {
                        'sale': sale,
                        'type': 'line',
                        'quantity': 1,
                        'product': product.products[0].id,
                        'unit': self.uom,
                        'unit_price': Decimal('50.95'),
                        'description': 'sale line',
                    }
                ])
                self.Sale.round_down_total([sale])
                # Previous roundoff line should be deleted.
                round_off_lines = self.SaleLine.search_count([
                    ('id', '=', round_off_line.id)
                ])
                self.assertEqual(round_off_lines, 0)

                # There should be a new roundoff line created
                round_off_lines = self.SaleLine.search([
                    ('is_round_off', '=', True)
                ])
                # There should only be one roundoff line.
                self.assertEqual(len(round_off_lines), 1)
                self.assertEqual(round_off_lines[0].amount, Decimal('-0.20'))
                self.assertEqual(sale.total_amount, 251)

                # Process sale
                self.Sale.quote([sale])
                self.Sale.confirm([sale])

                # Processing sale which doesn't have round off account and
                # has a roundoff line, raises UserError.
                self.assertRaises(UserError, self.Sale.process, [sale])

                # Set round down account.
                self.saleConfiguration = self.SaleConfiguration.create([{
                    'round_down_account':
                        self._get_account_by_kind('revenue').id,
                }])
                self.Sale.process([sale])

                invoice, = self.Invoice.search([
                    ('sales', 'in', [sale.id])
                ])
                # There should be an invoice created from the processed sale
                self.assertEqual(invoice.total_amount, 251)

    def test_1155_round_off_case_2(self):
        """
        Process sale multple times and ensure only 1 invoice is created.
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            product, = self.Product.create([{
                'name': 'Test product',
                'list_price': 200,
                'cost_price': 200,
                'default_uom': self.uom,
                'salable': True,
                'sale_uom': self.uom,
                'account_expense': self._get_account_by_kind('expense').id,
                'account_revenue': self._get_account_by_kind('revenue').id,
                'products': [('create', [
                    {}
                ])]
            }])
            sale, = self.Sale.create([{
                'reference': 'Test Sale 1',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'company': self.company.id,
                'lines': [('create', [
                    {
                        'type': 'line',
                        'quantity': 1,
                        'product': product.products[0].id,
                        'unit': self.uom,
                        'unit_price': Decimal(200.25),
                        'description': 'sale line',
                    }
                ])]
            }])

            with Transaction().set_context(company=self.company.id):
                self.Sale.round_down_total([sale])
                self.assertEqual(len(sale.lines), 2)

                round_off_line, = self.SaleLine.search([
                    ('is_round_off', '=', True)
                ])

                # There should be a new line of type 'roundoff'
                self.assertIsNotNone(round_off_line)

                # Total order price 200.25 should have been rounded down to 200
                self.assertEqual(sale.total_amount, 200)
                # Difference after rounding down should be created as
                # roundoff line.
                self.assertEqual(round_off_line.unit_price, 0.25)
                self.assertEqual(round_off_line.quantity, -1)
                self.assertEqual(round_off_line.amount, -0.25)

                self.SaleLine.create([
                    {
                        'sale': sale,
                        'type': 'line',
                        'quantity': 1,
                        'product': product.products[0].id,
                        'unit': self.uom,
                        'unit_price': Decimal('50.95'),
                        'description': 'sale line',
                    }
                ])
                self.Sale.round_down_total([sale])
                # Previous roundoff line should be deleted.
                round_off_lines = self.SaleLine.search_count([
                    ('id', '=', round_off_line.id)
                ])
                self.assertEqual(round_off_lines, 0)

                # There should be a new roundoff line created
                round_off_lines = self.SaleLine.search([
                    ('is_round_off', '=', True)
                ])
                # There should only be one roundoff line.
                self.assertEqual(len(round_off_lines), 1)
                self.assertEqual(round_off_lines[0].amount, Decimal('-0.20'))
                self.assertEqual(sale.total_amount, 251)

                # Process sale
                self.Sale.quote([sale])
                self.Sale.confirm([sale])

                # Processing sale which doesn't have round off account and
                # has a roundoff line, raises UserError.
                self.assertRaises(UserError, self.Sale.process, [sale])

                # Set round down account.
                self.saleConfiguration = self.SaleConfiguration.create([{
                    'round_down_account':
                        self._get_account_by_kind('revenue').id,
                }])
                self.Sale.process([sale])
                self.Sale.process([sale])
                self.Sale.process([sale])
                self.Sale.process([sale])

                invoices = self.Invoice.search([])
                self.assertEqual(len(invoices), 1)
                self.assertEqual(invoices[0].total_amount, 251)

    def test_1156_round_off_case_3(self):
        """
        Process sale and cancel it's invoice. Then process sale again and
        ensure that a new invoice is created
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            product, = self.Product.create([{
                'name': 'Test product',
                'list_price': 200,
                'cost_price': 200,
                'default_uom': self.uom,
                'salable': True,
                'sale_uom': self.uom,
                'account_expense': self._get_account_by_kind('expense').id,
                'account_revenue': self._get_account_by_kind('revenue').id,
                'products': [('create', [
                    {}
                ])]
            }])
            sale, = self.Sale.create([{
                'reference': 'Test Sale 1',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'company': self.company.id,
                'lines': [('create', [
                    {
                        'type': 'line',
                        'quantity': 1,
                        'product': product.products[0].id,
                        'unit': self.uom,
                        'unit_price': Decimal(200.25),
                        'description': 'sale line',
                    }
                ])]
            }])

            with Transaction().set_context(company=self.company.id):
                self.Sale.round_down_total([sale])
                self.assertEqual(len(sale.lines), 2)

                round_off_line, = self.SaleLine.search([
                    ('is_round_off', '=', True)
                ])

                # There should be a new line of type 'roundoff'
                self.assertIsNotNone(round_off_line)

                # Total order price 200.25 should have been rounded down to 200
                self.assertEqual(sale.total_amount, 200)
                # Difference after rounding down should be created as
                # roundoff line.
                self.assertEqual(round_off_line.unit_price, 0.25)
                self.assertEqual(round_off_line.quantity, -1)
                self.assertEqual(round_off_line.amount, -0.25)

                self.SaleLine.create([
                    {
                        'sale': sale,
                        'type': 'line',
                        'quantity': 1,
                        'product': product.products[0].id,
                        'unit': self.uom,
                        'unit_price': Decimal('50.95'),
                        'description': 'sale line',
                    }
                ])
                self.Sale.round_down_total([sale])
                # Previous roundoff line should be deleted.
                round_off_lines = self.SaleLine.search_count([
                    ('id', '=', round_off_line.id)
                ])
                self.assertEqual(round_off_lines, 0)

                # There should be a new roundoff line created
                round_off_lines = self.SaleLine.search([
                    ('is_round_off', '=', True)
                ])
                # There should only be one roundoff line.
                self.assertEqual(len(round_off_lines), 1)
                self.assertEqual(round_off_lines[0].amount, Decimal('-0.20'))
                self.assertEqual(sale.total_amount, 251)

                # Process sale
                self.Sale.quote([sale])
                self.Sale.confirm([sale])

                # Processing sale which doesn't have round off account and
                # has a roundoff line, raises UserError.
                self.assertRaises(UserError, self.Sale.process, [sale])

                # Set round down account.
                self.saleConfiguration = self.SaleConfiguration.create([{
                    'round_down_account':
                        self._get_account_by_kind('revenue').id,
                }])
                self.Sale.process([sale])
                self.Sale.process([sale])
                self.Sale.process([sale])
                self.Sale.process([sale])

                invoices = self.Invoice.search([])
                self.assertEqual(len(invoices), 1)
                self.assertEqual(invoices[0].total_amount, 251)

                self.Invoice.cancel(invoices)

                self.Sale.process([sale])

                invoices = self.Invoice.search([])
                self.assertEqual(len(invoices), 1)
                self.assertEqual(invoices[0].total_amount, 251)

                self.Sale.process([sale])
                self.Sale.process([sale])
                self.Sale.process([sale])

                invoices = self.Invoice.search([])
                self.assertEqual(len(invoices), 1)
                self.assertEqual(invoices[0].total_amount, 251)

                # Manually call get invoice line for round off line and
                # check if credit note is not created
                round_off_lines[0].invoice_lines = None
                round_off_lines[0].save()
                invoice_line = round_off_lines[0].get_invoice_line(
                    'out_credit_note')
                self.assertEqual(invoice_line, [])

    def test_1157_round_off_case_4(self):
        """
        Negative amount on sale
        """
        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()
            product, = self.Product.create([{
                'name': 'Test product',
                'list_price': 200,
                'cost_price': 200,
                'default_uom': self.uom,
                'salable': True,
                'sale_uom': self.uom,
                'account_expense': self._get_account_by_kind('expense').id,
                'account_revenue': self._get_account_by_kind('revenue').id,
                'products': [('create', [
                    {}
                ])]
            }])
            sale, = self.Sale.create([{
                'reference': 'Test Sale 1',
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'company': self.company.id,
                'lines': [('create', [
                    {
                        'type': 'line',
                        'quantity': -1,
                        'product': product.products[0].id,
                        'unit': self.uom,
                        'unit_price': Decimal(200.25),
                        'description': 'sale line',
                    }
                ])]
            }])

            with Transaction().set_context(company=self.company.id):
                self.Sale.round_down_total([sale])
                self.assertEqual(len(sale.lines), 2)

                round_off_line, = self.SaleLine.search([
                    ('is_round_off', '=', True)
                ])

                # There should be a new line of type 'roundoff'
                self.assertIsNotNone(round_off_line)

                # Total order price 200.25 should have been rounded down to 200
                self.assertEqual(sale.total_amount, -201)
                # Difference after rounding down should be created as
                # roundoff line.
                self.assertEqual(round_off_line.unit_price, 0.75)
                self.assertEqual(round_off_line.quantity, -1)
                self.assertEqual(round_off_line.amount, -0.75)

                # Process sale
                self.Sale.quote([sale])
                self.Sale.confirm([sale])

                # Set round down account.
                self.saleConfiguration = self.SaleConfiguration.create([{
                    'round_down_account':
                        self._get_account_by_kind('revenue').id,
                }])
                self.Sale.process([sale])

                invoices = self.Invoice.search([])
                self.assertEqual(len(invoices), 1)
                self.assertEqual(invoices[0].type, 'out_credit_note')
                self.assertEqual(invoices[0].total_amount, 201)

    def test_1160_sale_stays_in_confirm_state_forever(self):
        """
        If a line is pickup with zero total, sale cannot be done.
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,
                'invoice_method': 'shipment',
                'shipment_method': 'order',
            }])
            self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': 2,
                'delivery_mode': 'pick_up',
                'unit': self.uom,
                'unit_price': Decimal('0'),
                'description': 'Picked Item',
                'product': self.product1.id
            }])

            with Transaction().set_context({'company': self.company.id}):
                # Quote, Confirm and Process the order
                self.Sale.quote([sale])
                self.Sale.confirm([sale])
                self.Sale.process([sale])

                self.assertEqual(len(sale.shipments), 1)
                self.assertEqual(len(sale.invoices), 1)
                self.assertEqual(sale.invoice_state, 'paid')
                self.assertEqual(sale.shipment_state, 'sent')

                self.assertEqual(sale.state, 'done')

    def test_1170_test_assign_pick_up_shipments(self):
        """
        If a line is pickup with zero total, sale cannot be done.
        """
        Date = POOL.get('ir.date')

        with Transaction().start(DB_NAME, USER, context=CONTEXT):
            self.setup_defaults()

            sale, = self.Sale.create([{
                'payment_term': self.payment_term,
                'currency': self.company.currency.id,
                'party': self.party.id,
                'invoice_address': self.party.addresses[0].id,
                'shipment_address': self.party.addresses[0].id,
                'sale_date': Date.today(),
                'company': self.company.id,
                'invoice_method': 'shipment',
                'shipment_method': 'order',
            }])
            self.SaleLine.create([{
                'sale': sale,
                'type': 'line',
                'quantity': 100,
                'delivery_mode': 'pick_up',
                'unit': self.uom,
                'unit_price': Decimal('100'),
                'description': 'Picked Item',
                'product': self.product1.id
            }])

            with Transaction().set_context({'company': self.company.id}):
                # Quote, Confirm and Process the order
                self.Sale.quote([sale])
                self.Sale.confirm([sale])
                with self.assertRaises(UserError):
                    self.Sale.process([sale])


def suite():
    """
    Define suite
    """
    test_suite = trytond.tests.test_tryton.suite()
    test_suite.addTests(
        unittest.TestLoader().loadTestsFromTestCase(TestSale)
    )
    return test_suite

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite())



# -*- coding: utf-8 -*-
"""
    sale.py

    :copyright: (c) 2014 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
from datetime import datetime, timedelta
from sql import Literal
from trytond.model import fields
from trytond.pool import Pool, PoolMeta
from trytond.transaction import Transaction
from trytond.rpc import RPC
from trytond.model import ModelView
from trytond.pyson import Eval, Bool, And
from trytond import backend
from math import floor
from decimal import Decimal


__metaclass__ = PoolMeta
__all__ = ["Sale", "SaleChannel", "SaleLine"]


class SaleConfiguration:
    'Sale Configuration'
    __name__ = 'sale.configuration'

    round_down_account = fields.Property(
        fields.Many2One('account.account', 'Round Down Account', required=True)
    )


class SaleChannel:
    __name__ = 'sale.channel'

    anonymous_customer = fields.Many2One(
        'party.party', "Anonymous Customer", required=True
    )

    # The warehouse from which order lines with ship will be shipped
    ship_from_warehouse = fields.Many2One(
        'stock.location', "Warehouse (Shipped Lines)",
        required=True, domain=[('type', '=', 'warehouse')],
    )

    delivery_mode = fields.Selection([
        ('pick_up', 'Pick Up'),
        ('ship', 'Ship'),
    ], 'Delivery Mode', required=True)

    @staticmethod
    def default_delivery_mode():
        return 'ship'


class Sale:
    __name__ = "sale.sale"

    @staticmethod
    def default_party():
        User = Pool().get('res.user')
        user = User(Transaction().user)
        if (
            'use_anonymous_customer' not in Transaction().context
        ):  # pragma: no cover
            return
        if user.current_channel and user.current_channel.anonymous_customer:
            return user.current_channel.anonymous_customer.id

    @classmethod
    def __setup__(cls):
        super(Sale, cls).__setup__()
        cls.__rpc__.update({
            'pos_add_product': RPC(instantiate=0, readonly=False),
            'pos_serialize': RPC(instantiate=0, readonly=True),
            'get_recent_sales': RPC(readonly=True),
        })
        cls.lines.context = {
            'current_sale_channel': Eval('channel'),
        }
        cls._buttons.update({
            'round_down_total': {
                'invisible': ~Eval('state').in_(['draft', 'quotation']),
            },
        })

    @classmethod
    @ModelView.button
    def round_down_total(cls, records):
        '''
        Round down total order price and add remaining amount as new sale line
        '''
        SaleLine = Pool().get('sale.line')

        sale_lines = []
        for record in records:
            # Check if there's already a roundoff line, remove and create new
            # if there is.
            round_off_line = SaleLine.search([
                ('sale', '=', record.id),
                ('is_round_off', '=', True),
            ])
            if round_off_line:
                SaleLine.delete(round_off_line)

            floored_total = floor(record.total_amount)
            amount_diff = record.total_amount - Decimal(floored_total)
            sale_lines.append({
                'sale': record,
                'is_round_off': True,
                'type': 'line',
                'quantity': -1,
                'unit_price': amount_diff,
                'description': 'Round Off'
            })

        SaleLine.create(
            [line for line in sale_lines if line['unit_price']]
        )

    @classmethod
    def get_recent_sales(cls):
        """
        Return sales of current channel, which were made within last 5 days
        and are in draft state. Sort by write_date or create_date of Sale and
        sale lines.
        """
        SaleLine = Pool().get('sale.line')

        context = Transaction().context
        date = (
            datetime.now() - timedelta(days=5)
        ).strftime('%Y-%m-%d %H:%M:%S')
        current_channel = context['current_channel']

        SaleTable = cls.__table__()
        SaleLineTable = SaleLine.__table__()

        cursor = Transaction().cursor
        query = SaleTable.join(
            SaleLineTable,
            condition=(SaleTable.id == SaleLineTable.sale)
        ).select(
            SaleTable.id,
            where=(
                (SaleTable.channel == Literal(current_channel)) &
                (SaleTable.state.in_([
                    'draft', 'quotation', 'confirmed', 'processing'
                ])) &
                (
                    (SaleTable.write_date >= Literal(date)) |
                    (SaleTable.create_date >= Literal(date))
                )
            ),
            order_by=(
                SaleLineTable.write_date.desc,
                SaleLineTable.create_date.desc,
                SaleTable.write_date.desc,
                SaleTable.create_date.desc
            )
        )
        cursor.execute(*query)
        ids = [x[0] for x in cursor.fetchall()]
        return [cls(id).serialize('recent_sales') for id in ids]

    def pos_find_sale_line_domain(self):
        """
        Return domain to find existing sale line for given product.
        """
        domain = [
            ('sale', '=', self.id),
        ]

        context = Transaction().context

        if 'product' in context:
            domain.append(('product', '=', context['product']))

        if 'delivery_mode' in context:
            domain.append(('delivery_mode', '=', context['delivery_mode']))

        return domain

    def pos_add_product(self, product_ids, quantity):
        """
        Add product to sale from POS.
        This method is for POS, to add multiple products to cart in single call
        """
        AccountTax = Pool().get('account.tax')
        SaleLine = Pool().get('sale.line')

        updated_lines = []
        for product_id in product_ids:
            Transaction().set_context(product=product_id)
            try:
                if 'sale_line' in Transaction().context:
                    sale_line = SaleLine(Transaction().context.get('sale_line'))
                else:
                    sale_line, = SaleLine.search(
                        self.pos_find_sale_line_domain()
                    )
            except ValueError:
                sale_line = None

            delivery_mode = Transaction().context.get(
                'delivery_mode', 'pick_up'
            )

            if sale_line:
                values = {
                    'product': sale_line.product.id,
                    '_parent_sale.currency': self.currency.id,
                    '_parent_sale.party': self.party.id,
                    '_parent_sale.price_list': (
                        self.price_list.id if self.price_list else None
                    ),
                    '_parent_sale.sale_date': self.sale_date,
                    '_parent_sale.channel': self.channel,
                    '_parent_sale.shipment_address': self.shipment_address,
                    'warehouse': self.warehouse,
                    '_parent_sale.warehouse': self.warehouse,
                    'unit': sale_line.unit.id,
                    'quantity': quantity,
                    'type': 'line',
                    'delivery_mode': delivery_mode,
                }

                # Update the values by triggering an onchange which should
                # fill missing vals
                values.update(SaleLine(**values).on_change_quantity())
                values.update(SaleLine(**values).on_change_delivery_mode())

                new_values = {}
                for key, value in values.iteritems():
                    if '.' in key:
                        continue
                    if key == 'taxes':
                        # Difficult to reach here unless taxes change when
                        # quantities change.
                        continue    # pragma: no cover
                    new_values[key] = value
                SaleLine.write([sale_line], new_values)
            else:
                values = {
                    'product': product_id,
                    '_parent_sale.currency': self.currency.id,
                    '_parent_sale.party': self.party.id,
                    '_parent_sale.price_list': (
                        self.price_list.id if self.price_list else None
                    ),
                    '_parent_sale.sale_date': self.sale_date,
                    '_parent_sale.channel': self.channel,
                    '_parent_sale.shipment_address': self.shipment_address,
                    'warehouse': self.warehouse,
                    '_parent_sale.warehouse': self.warehouse,
                    'sale': self.id,
                    'type': 'line',
                    'quantity': quantity,
                    'unit': None,
                    'description': None,
                    'delivery_mode': delivery_mode,
                }
                values.update(SaleLine(**values).on_change_product())
                values.update(SaleLine(**values).on_change_quantity())
                values.update(SaleLine(**values).on_change_delivery_mode())
                new_values = {}
                for key, value in values.iteritems():
                    if '.' in key:
                        continue
                    if key == 'taxes':
                        continue
                    new_values[key] = value
                sale_line = SaleLine.create([new_values])[0]

            updated_lines.append(sale_line.id)
            if 'taxes' in values:
                sale_line.taxes = AccountTax.browse(values['taxes'])
                sale_line.save()

        # Now that the sale line is built, return a serializable response
        # which ensures that the client does not have to call again.
        res = {
            'sale': self.serialize('pos'),
            'updated_lines': updated_lines,
        }
        return res

    def pos_serialize(self):
        """
        Serialize sale for pos
        """
        return self.serialize('pos')

    def serialize(self, purpose=None):
        """
        Serialize with information needed for POS
        """
        Address = Pool().get('party.address')

        invoice_address = Address.search([
            ('party', '=', self.party.id),
            ('invoice', '=', True)
        ], limit=1)

        shipment_address = Address.search([
            ('party', '=', self.party.id),
            ('delivery', '=', True)
        ], limit=1)

        if purpose == 'pos':
            invoice_address = self.invoice_address or \
                invoice_address[0] if invoice_address else None
            shipment_address = self.shipment_address or \
                shipment_address[0] if shipment_address else None
            return {
                'party': self.party.id,
                'total_amount': self.total_amount,
                'untaxed_amount': self.untaxed_amount,
                'tax_amount': self.tax_amount,
                'comment': self.comment,
                'state': self.state,
                'invoice_address': invoice_address and
                    invoice_address.serialize(purpose),
                'shipment_address': shipment_address and
                    shipment_address.serialize(purpose),
                'lines': [line.serialize(purpose) for line in self.lines],
                'reference': self.reference,
            }
        elif purpose == 'recent_sales':
            return {
                'id': self.id,
                'party': {
                    'id': self.party.id,
                    'name': self.party.name,
                },
                'total_amount': self.total_amount,
                'create_date': self.create_date,
                'state': self.state,
                'reference': self.reference,
            }
        elif hasattr(super(Sale, self), 'serialize'):
            return super(SaleLine, self).serialize(purpose)  # pragma: no cover

    def _group_shipment_key(self, moves, move):
        """
        This method returns a key based on which Tryton creates shipments
        for a given sale order. By default Tryton uses the planned_date for the
        delivery and warehouse to separate shipments.

        We use the same functionality to split the shipments for items being
        picked up and delivered. This is later used to auto proceed and finish
        the shipping of the picked up products.

        :param moves: A list of all moves
        :param move: move is a tuple of line id and a move
        """
        SaleLine = Pool().get('sale.line')

        line = SaleLine(move[0])
        rv = super(Sale, self)._group_shipment_key(moves, move)
        return rv + (('delivery_mode', line.delivery_mode),)

    def create_shipment(self, shipment_type):
        """
        This method creates the shipments for the given sale order.

        This implementation inspects the order lines to look for lines which
        are expected to be picked up instantly and the shipment created for
        pick_up is automatically processed all the way through.
        """
        pool = Pool()

        shipments = super(Sale, self).create_shipment(shipment_type)

        if self.shipment_method == 'manual':
            # shipments will be None but for future return the value
            # returned by the super function
            return shipments

        if not shipments:
            return shipments

        picked_up_shipments = filter(
            lambda s: s.delivery_mode == 'pick_up', shipments
        )

        if shipment_type == 'out':
            Shipment = pool.get('stock.shipment.out')

            with Transaction().set_user(0, set_context=True):
                # If we are going to "process" a shipment, it is
                # equivalent to sale being processed.
                #
                # Doing this here helps in an edge case where the
                # sale total is 0. When a shipment is "Done" it
                # tries to recprocess the order state, but
                # usually this happens after sale is in
                # processing state. Since we push things through in the
                # next few lines, that call happens when the sale is in
                # confirmed state and there is no transition from
                # Confirmed to Done.
                self.state = 'processing'
                self.save()
                # Assign and complete the shipments
                if not Shipment.assign_try(picked_up_shipments):
                    draft_moves = filter(
                        lambda m: m.state == 'draft',
                        [m for s in picked_up_shipments for m in s.outgoing_moves]  # noqa
                    )
                    products_out_of_stock = [
                        m.product.rec_name for m in draft_moves
                    ]
                    self.raise_user_error(
                        "Order cannot be processed as the following items are "
                        "out of stock:\n" + "\n".join(products_out_of_stock)
                    )
                Shipment.pack(picked_up_shipments)
                Shipment.done(picked_up_shipments)
        elif shipment_type == 'return':
            Shipment = pool.get('stock.shipment.out.return')
            with Transaction().set_user(0, set_context=True):
                Shipment.receive(picked_up_shipments)
                Shipment.done(picked_up_shipments)

        # Finally return the value the super function returned, but after
        # reloading the active records.
        return Shipment.browse(map(int, shipments))

    def create_invoice(self, invoice_type):
        """
        Sale creates draft invoices. But if the invoices are created from
        shipments, then they should be automatically opened
        """
        Invoice = Pool().get('account.invoice')

        invoice = super(Sale, self).create_invoice(invoice_type)

        if not invoice:
            return invoice

        if self.invoice_method == 'shipment' and invoice_type == 'out_invoice':
            # Invoices created from shipment can be automatically opened
            # for payment.
            Invoice.post([invoice])

        return invoice


class SaleLine:
    __name__ = 'sale.line'

    is_round_off = fields.Boolean('Round Off', readonly=True)

    delivery_mode = fields.Selection([
        (None, ''),
        ('pick_up', 'Pick Up'),
        ('ship', 'Ship'),
    ], 'Delivery Mode', states={
        'invisible': Eval('type') != 'line',
        'required': And(
            Eval('type') == 'line',
            Bool(Eval('product_type_is_goods'))
        )
    }, depends=['type', 'product_type_is_goods'])

    product_type_is_goods = fields.Function(
        fields.Boolean('Product Type is Goods?'), 'get_product_type_is_goods'
    )

    @classmethod
    def __register__(cls, module_name):
        super(SaleLine, cls).__register__(module_name)

        TableHandler = backend.get('TableHandler')
        cursor = Transaction().cursor

        table = TableHandler(cursor, cls, module_name)

        table.not_null_action('delivery_mode', action='remove')

    @classmethod
    def __setup__(cls):
        super(SaleLine, cls).__setup__()

        # Hide product and unit fields.
        cls.product.states['invisible'] |= Bool(Eval('is_round_off'))
        cls.unit.states['invisible'] |= Bool(Eval('is_round_off'))
        cls.delivery_mode.states['invisible'] |= Bool(Eval('is_round_off'))
        cls.product.depends.insert(0, 'is_round_off')
        cls.unit.depends.insert(0, 'is_round_off')

    @fields.depends(
        'product', 'unit', 'quantity', '_parent_sale.party',
        '_parent_sale.currency', '_parent_sale.sale_date', 'delivery_mode',
        '_parent_sale.channel', '_parent_sale.shipment_address',
        'warehouse', '_parent_sale.warehouse'
    )
    def on_change_delivery_mode(self):
        """
        This method can be overridden by downstream modules to make changes
        according to delivery mode. Like change taxes according to delivery
        mode.
        """
        return {}

    @staticmethod
    def default_is_round_off():
        return False

    def get_invoice_line(self, invoice_type):
        SaleConfiguration = Pool().get('sale.configuration')
        InvoiceLine = Pool().get('account.invoice.line')

        if not self.is_round_off:
            return super(SaleLine, self).get_invoice_line(invoice_type)

        # The line is a round off line and apply the logic here.

        # Check if the invoice line already exists for the sale line
        # If yes, then no line needs to be created
        # XXX: What if the invoice was cancelled ?
        if self.invoice_lines:
            return []

        round_down_account = SaleConfiguration(1).round_down_account
        if not round_down_account:
            self.raise_user_error(
                '''Set round down account from Sale Configuration to
                add round off line'''
            )

        invoice_line = InvoiceLine()
        invoice_line.origin = self
        invoice_line.account = round_down_account
        invoice_line.unit_price = self.unit_price
        invoice_line.description = self.description

        # For positive sales transactions (where the order is effectively
        # a positive total), the round_down is applied on out_invoice
        # and if overall order total is negative, then the round_down is
        # tied to credit note.
        if self.sale.total_amount >= Decimal('0'):
            if invoice_type == 'out_credit_note':
                # positive order looking for credit note
                return []
            invoice_line.quantity = self.quantity
        else:
            if invoice_type == 'out_invoice':
                # negative order looking for invoice
                return []
            invoice_line.quantity = -self.quantity

        return [invoice_line]

    @staticmethod
    def default_delivery_mode():
        Channel = Pool().get('sale.channel')
        User = Pool().get('res.user')

        user = User(Transaction().user)
        sale_channel = user.current_channel
        if Transaction().context.get('current_sale_channel'):
            sale_channel = Channel(
                Transaction().context.get('current_sale_channel')
            )
        return sale_channel and sale_channel.delivery_mode

    def get_warehouse(self, name):
        """
        Return the warehouse from the channel for orders being picked up and the
        backorder warehouse for orders with ship.
        """
        if self.delivery_mode == 'ship':
            return self.sale.channel.ship_from_warehouse.id
        return super(SaleLine, self).get_warehouse(name)

    def serialize(self, purpose=None):
        """
        Serialize for the purpose of POS
        """
        if purpose == 'pos':
            return {
                'id': self.id,
                'description': self.description,
                'product': self.product and {
                    'id': self.product.id,
                    'code': self.product.code,
                    'rec_name': self.product.rec_name,
                    'default_image': self.product.default_image and
                                    self.product.default_image.id,
                },
                'unit': self.unit and {
                    'id': self.unit.id,
                    'rec_name': self.unit.rec_name,
                },
                'unit_price': self.unit_price,
                'quantity': self.quantity,
                'amount': self.amount,
                'delivery_mode': self.delivery_mode
            }
        elif hasattr(super(SaleLine, self), 'serialize'):
            return super(SaleLine, self).serialize(purpose)  # pragma: no cover

    def get_product_type_is_goods(self, name):
        """
        Return True if product is of type goods
        """
        if self.product and self.product.type == 'goods':
            return True
        return False



# -*- coding: utf-8 -*-
"""
    shipment.py

    :copyright: (c) 2014 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
from trytond.model import fields
from trytond.pool import PoolMeta

__metaclass__ = PoolMeta
__all__ = ['ShipmentOut', 'ShipmentOutReturn']


class ShipmentOut:
    __name__ = 'stock.shipment.out'

    delivery_mode = fields.Selection([
        ('pick_up', 'Pick Up'),
        ('ship', 'Ship'),
    ], 'Delivery Mode', required=True, readonly=True)

    @staticmethod
    def default_delivery_mode():
        return 'ship'


class ShipmentOutReturn:
    __name__ = 'stock.shipment.out.return'

    # XXX: Not sure if pick_up is the right word here ?
    # My intention is to indicate something like the customer walked into
    # the store and returned this item.
    #
    # While ship indicates that the customer intends to ship the item back
    # to us later by mail or something like that.
    delivery_mode = fields.Selection([
        ('pick_up', 'Pick Up'),
        ('ship', 'Ship'),
    ], 'Delivery Mode', required=True, readonly=True)

    @staticmethod
    def default_delivery_mode():
        return 'ship'



# -*- coding: utf-8 -*-
"""
    tests/test_address.py

    :copyright: (C) 2014 by Openlabs Technologies & Consulting (P) Limited
    :license: BSD, see LICENSE for more details.
"""
import sys
import os
import unittest

import trytond.tests.test_tryton
from trytond.tests.test_tryton import POOL, USER, DB_NAME, CONTEXT
from trytond.transaction import Transaction

DIR = os.path.abspath(os.path.normpath(os.path.join(
    __file__, '..', '..', '..', '..', '..', 'trytond'
)))
if os.path.isdir(DIR):
    sys.path.insert(0, os.path.dirname(DIR))


class TestAddress(unittest.TestCase):
    '''
    Address Test Case for lie-nielsen module.
    '''

    def setup_defaults(self):
        """
        Setup Defaults
        """
        self.Party = POOL.get('party.party')
        self.Address = POOL.get('party.address')

        with Transaction().set_context(company=None):
            self.party1, = self.Party.create([{
                'name': 'Jon Doe'
            }])
            self.address, = self.Address.create([{
                'party': self.party1,
                'name': 'Jon Doe\'s Address'
            }])

    def test_0010_test_address_serialization(self):
        """
        Test address serialization
        """
        with Transaction().start(DB_NAME, USER, CONTEXT):
            self.setup_defaults()
            self.address.serialize('pos')
            self.address.serialize()


def suite():
    """
    Define suite
    """
    test_suite = trytond.tests.test_tryton.suite()
    test_suite.addTests(
        unittest.TestLoader().loadTestsFromTestCase(TestAddress)
    )
    return test_suite

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite())



#!/usr/bin/env python
import re
import os
import sys
import time
import unittest
import ConfigParser
from setuptools import setup, Command


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class SQLiteTest(Command):
    """
    Run the tests on SQLite
    """
    description = "Run tests on SQLite"

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.environ['TRYTOND_DATABASE_URI'] = 'sqlite://'
        os.environ['DB_NAME'] = ':memory:'

        from tests import suite
        test_result = unittest.TextTestRunner(verbosity=3).run(suite())

        if test_result.wasSuccessful():
            sys.exit(0)
        sys.exit(-1)


class PostgresTest(Command):
    """
    Run the tests on Postgres.
    """
    description = "Run tests on Postgresql"

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.environ['TRYTOND_DATABASE_URI'] = 'postgresql://'

        os.environ['DB_NAME'] = 'test_' + str(int(time.time()))

        from tests import suite
        test_result = unittest.TextTestRunner(verbosity=3).run(suite())

        if test_result.wasSuccessful():
            sys.exit(0)
        sys.exit(-1)


config = ConfigParser.ConfigParser()
config.readfp(open('tryton.cfg'))
info = dict(config.items('tryton'))
for key in ('depends', 'extras_depend', 'xml'):
    if key in info:
        info[key] = info[key].strip().splitlines()
major_version, minor_version, _ = info.get('version', '0.0.1').split('.', 2)
major_version = int(major_version)
minor_version = int(minor_version)

requires = []

MODULE2PREFIX = {
    'sale_channel': 'openlabs',
}

MODULE = "pos"
PREFIX = "openlabs"
for dep in info.get('depends', []):
    if not re.match(r'(ir|res|webdav)(\W|$)', dep):
        requires.append(
            '%s_%s >= %s.%s, < %s.%s' % (
                MODULE2PREFIX.get(dep, 'trytond'), dep,
                major_version, minor_version, major_version,
                minor_version + 1
            )
        )
requires.append(
    'trytond >= %s.%s, < %s.%s' % (
        major_version, minor_version, major_version, minor_version + 1
    )
)
setup(
    name='%s_%s' % (PREFIX, MODULE),
    version=info.get('version', '0.0.1'),
    description="Tryton POS helper module",
    author="Openlabs Technologies and Consulting (P) Ltd.",
    author_email='info@openlabs.co.in',
    url='http://www.openlabs.co.in/',
    package_dir={'trytond.modules.%s' % MODULE: '.'},
    packages=[
        'trytond.modules.%s' % MODULE,
        'trytond.modules.%s.tests' % MODULE,
    ],
    package_data={
        'trytond.modules.%s' % MODULE: info.get('xml', []) +
        info.get('translation', []) +
        ['tryton.cfg', 'locale/*.po', 'tests/*.rst', 'reports/*.odt'] +
        ['view/*.xml'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Tryton',
        'Topic :: Office/Business',
    ],
    license='GPL-3',
    install_requires=requires,
    tests_require=['pycountry'],
    zip_safe=False,
    entry_points="""
    [trytond.modules]
    %s = trytond.modules.%s
    """ % (MODULE, MODULE),
    test_suite='tests',
    test_loader='trytond.test_loader:Loader',
    cmdclass={
        'test': SQLiteTest,
        'test_on_postgres': PostgresTest,
    }
)
